! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module grismdetection_mod

  use cal
  use dal
  use Errorhandling
  use omallocation_mod
  use omconstants_mod
  use omutility_mod

  use OmSourceParam_mod    ! Defines structures for storing source parameters
  use parameters_mod       ! For computing source moments
  use OmSmoothImage_mod    ! For smoothing image
 

implicit none 

 

!*********************************************
! Source-detection algorithm for grism data
! 
!
!********************************************

contains
Subroutine GrismDetection(sourceList, noSources, image, levelImage)
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list
    integer(int16), intent(out) :: noSources

    real(single), pointer :: image(:, :)
    integer(int16), pointer :: levelimage(:, :)


    integer(int32), pointer :: sortIndex(:) => null()
    integer :: i0, j0, i, j, j1, k, l, n, il, iu, jl, ju, n1, n2, n3, nu
    integer :: il1, iu1, il2, iu2
    integer :: nx, ny, minNo
    integer :: minN, nl, nc, nr, lwidth, cwidth, rwidth
    real(single) :: x, xl, xu, dx, pa, xmin, xmax
    real(single) :: fl, fc, fr
  
  
    noSources = 0

    write(errstr,'(a)') "Searching for first-order spectra"
    call message(AppMsg, Verbose)
 
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx = 6 / binx + binx ! Central region width   
    


    i0 = dx + 1
    nu = ny
    nl = max(0, nu - 200 / binx)
    minN = nint( real(nu - nl + 1) * 0.75)
 
    do i = 0, nx
        if (i0 .ge. nx) exit
        
        
        
        do k = 0, 6
            cwidth = dx + k
            il = i0 - cwidth
            if (il .lt. 0) cycle
            iu = i0 + cwidth
            if (iu .gt. nx) exit
            iu1 = il - 2
            il1 = iu1 - 5
            iu1 = max(0, iu1)
            il1 = max(0, il1)
   
            il2 = iu + 2
            iu2 = il2 + 2
            il2 = min(nx, il2)
            iu2 = min(nx, iu2)
         

            n1 = 0
           
            nc = iu - il + 1
            if (nc .lt. 3) cycle
            nl = iu1 - il1 + 1
            if (nl .lt. 3) cycle
          
            nr = iu2 - il2 + 1
            if (nr .lt. 3) cycle
          
            n = 0  ! Number of central pixels above local background
             
            do j = nu, nl, -1
                if ( any(image(il : iu, j) .lt. 0.0)) cycle   ! Skip negative pixels
                if ( any(image(il1 : iu1, j) .lt. 0.0)) cycle   ! Skip negative pixels
                if ( any(image(il2 : iu2, j) .lt. 0.0)) cycle   ! Skip negative pixels

                fc = sum(image(il : iu, j)) / real(nc)      ! The central averaged counts
                fl = sum(image(il1 : iu1, j)) / real(nl)    ! The left-background averaged counts
                fr = sum(image(il2 : iu2, j)) / real(nr)    ! The right-background averaged counts
                if (fc .gt. fl .and. fc .gt. fr) n = n + 1  ! Number of pixels above the local background  
               
            end do
            if (n .gt. minN) then ! Possible spectum
                noSources = noSources + 1
                do j = il, iu
                    do l = jl, ju
                        levelimage(j, l) = noSources
                    end do
                end do          
                i0 = i0 + dx  
                exit
            end if
        end do
        i0 = i0 + dx
    end do
    write(errstr,'(a, i5)') "Number of possible first-order spectra found=", noSources

    call message(AppMsg, Verbose)
 
  
 
end Subroutine GrismDetection



end module grismdetection_mod
