! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module optimize_mod
use dal
use omutility_mod
use omnumerics_mod
use omconstants_mod

integer, private, pointer :: index(:) => null()

   
public


type fitParameters
    
    real(double) :: x(20)   
    real(double) :: fMax

end type fitParameters


contains

!*************************************************************
! J, Canto, S. Curiel and E. Martinez-Gomez, 2009, 
! Astonomy and Astrophysics, 501, 1259.
!
! "A simple algorithm for optmization and model fitting:
! AGA (asexual genetic algorithm)"
!
!
!************************************************************

!**********************************************************************
!
!
!
!
!**********************************************************************
subroutine optimize(nToFit, fitNumber, x, &
           lBound, uBound, n0In, ngenIn, toleranceIn, functn, fMax, ngenout)
    integer, intent(in) :: nToFit
    integer, intent(in) :: fitNumber(:)
    real(double), dimension(:) :: x, lBound, uBound
    integer, intent(in), optional :: n0In, ngenIn
    real(double), intent(in), optional  :: toleranceIn
    real(double), intent(out) :: fMax
    integer, intent(out), optional :: ngenOut  

interface

    !*********************************************************
    ! The function to be minimised
    !*********************************************************
    function functn(x)
    use dal
    
       
    real (double) :: x(:)
    real(double) :: functn
        
    end  Function functn
end interface
    integer :: i, iOld, j, k, n0, ngen, n1, n2, il, iu, nits
    type(fitParameters), pointer :: xFitN0(:) => null()
    type(fitParameters), pointer :: xFitN1(:) => null()
    
    real(double), pointer, dimension(:) :: x0, width => null(), halfWidth => null()
    real(double), parameter :: p = 0.9330_double
    real(double) :: fMaxOld, tolerance
  
    call random_seed()
 

!write(*, *) "optimize_mod.f90 line 146, nterms=", nterms, " nToFit=", nToFit, " size(x)=", size(x)
!write(*, *) "fitnumbers"
!do i = 1, nToFit
!    write(*, *) i, ") ", fitNumber(i), " lbound=", lbound(i), " ubound=", ubound(i)
!end do

    if(present(n0In)) then
        n0 = n0In
    else
        n0 = 100
    end if

    if(present(nGenIn)) then
        nGen = nGenIn
    else
        nGen = 100
    end if

    if(present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 1e-6
    end if



    n1 = n0 / 10
    n2 = n0 / n1 - 1

    allocate(xFitN0(1 : n0))

!    do i = 1, n0
!        xFitN0(i)%x => allocateDouble(1, nToFit, "x")
!    end do

    allocate(xFitN1(1 : n1))
!    do i = 1, n1
!        xFitN1(i)%x => allocateDouble(1, nToFit, "x")
!    end do

    
    index => allocateInt32(1, n0, "index")


    width => allocateDouble(1, nToFit, "width")
    halfWidth => allocateDouble(1, nToFit, "halfwidth")

    x0 => allocateDouble(1, nToFit, "x0")
   


    !************************************************************
    !
    ! Construct a random initial population
    !
    !************************************************************
    do i = 1, nToFit

        j = fitNumber( i )
         
        width(i) = uBound(i)  - lbound(i)
        halfWidth(i) = width(i) / 2.0_double
        x0(i) = lbound(i) + halfWidth(i)
 
    end do

    call population(nToFit, fitNumber, x, x0, width, halfWidth, lBound, uBound, xFitN0, functn)

  
  

    fmaxOld = -1e10
    nits = 0
    iOld = 1
    do i = 1, nGen ! The iteration number
        nits = nits + 1

          

        !*************************************************************
        !
        ! Test for convergence
        !
        !************************************************************
        fmax = maxval(xFitn0%fmax)

       
 
        if( (i - iOld)  == 20) then
            
            if( fMax > fmaxOld) then

                if( abs(fMax - fmaxOld) / max(0.00000000001_double, abs(fMaxOld)) .le. tolerance) then
!                    write(*, *) "optimize- exiting after ", i, " generations, fmax=", fmax
                    exit
                end if
            end if
            iOld = i
            fmaxOld = fMax
        end if
          

        !***********************************************************
        !
        ! Select the n1 highest function values from  xFitn0
        !
        !***********************************************************
        call selectIndividuals(n0, xFitN0, n1, xFitn1) 

!        write(*, *) "Line 257 Iteration ", i, " max func value=", xFitn1(n1)%fmax, &
!            " x=", xfitn1(n1)%x(1 : nToFit)




        !**************************************************************
        ! Construct a new population using the individuals in xFitN1
        ! and store in xFitn2
        !***************************************************************
    
        
        !******************************************************
        !
        ! Compute the new widths
        !
        !******************************************************

        do j = 1, nToFit
            width(j) = (uBound(j)  - lbound(j)) * (p ** (i - 1))
            halfWidth(j) = width(j) / 2.0_double
        end do


        !*************************************************************
        !
        ! Keep a clone of the parent
        !
        !*************************************************************

        xFitn0(1 : n1) = xfitN1 
    
        do j = 1, n1         
            do k = 1, nToFit
                x0(k) = xFitN1(j)%x(k) 
                
            end do

            il = n1 + (j - 1) * n2 + 1
            iu = n1 + j * n2 

            call population(nToFit, fitNumber, x, x0, width, halfWidth, lBound, uBound, xFitn0(il : iu), functn )         

        end do





        
    end do

 
!write(*, *) "Line 311 max func value=", xFitn1(n1)%fmax, &
!            " x=", xfitn1(n1)%x(1 : nToFit)



    do i = 1, nToFit
        x(fitNumber(i))= xfitn1(n1)%x(i)
    end do
 
    fmax = xFitn1(n1)%fmax
    
    deallocate(xFitN0, xFitN1, width, halfWidth, x0, index)






!    write(*, *) "Optimez1- Fmax=", fmax, " number of generations=", nits
!    write(*, *) "new computed value=", functn(x)
!    do i =1, nterms
!        write(*, *) i, ") x=", x(i)     
!    end do

    if(present(ngenout)) then


        ngenout = nits
    end if


end subroutine optimize
!******************************************************************
!
!
!*******************************************************************
subroutine population(nToFit, fitNumber, x, x0, width, halfWidth, bl, bu, xFit, functn)
    integer, intent(in) :: nToFit, fitNumber(:)
    real(double), dimension(:) :: x, x0, halfWidth, width, bl, bu
    type(fitParameters) :: xFit(:)

interface

    !*********************************************************
    ! The function to be minimised
    !*********************************************************
    function functn(x)
    use dal
    
       
    real (double) :: x(:)
    real(double) :: functn
         
    end  Function functn
end interface


    integer :: i, j
    real(double) :: value

    do i = 1, size(xFit)
        do j = 1, nToFit
            call random_number(value)
            xFit(i)%x(j) = x0(j) - halfWidth(j) + value * width(j)
            xFit(i)%x(j) = max(bl(j), xFit(i)%x(j))
            xFit(i)%x(j) = min(bu(j), xFit(i)%x(j))
            x( fitNumber(j) ) = xFit(i)%x(j)

        end do 
        xFit(i)%fmax = functn( x ) 
    end do

!write(*, *) "Population- fmax=", xFit(size(xfit))%fMax, " maxval=", maxval(xfit%fmax), &
!" x=", xFit(size(xfit))%x, " x0=", x0


end subroutine population

!******************************************************************
!
!
!
!*******************************************************************
subroutine computeFitness(x, nToFit, fitNumber, xFit, functn)
    integer, intent(in) :: nToFit, fitNumber(:)
    real(double) :: x(:)
    type(fitParameters) :: xFit(:)

interface

    !*********************************************************
    ! The function to be minimised
    !*********************************************************
    function functn(x)
    use dal
    
       
    real (double) :: x(:)
    real(double) :: functn
         
    end  Function functn
end interface
    integer :: i, j

    !***********************************************************
    !
    ! Compute the function values for each set of parameters
    !
    !************************************************************
    do i = 1, size(xFit)
        do j = 1, nToFit
            x( fitNumber(j) ) = xFit(i)%x(j)
        end do
          

        xFit(i)%fmax = functn( x )         
    end do


end subroutine computeFitness
!************************************************************
! Select the subset n1 of the individuals with the highest
! fitness from xFitn0 and store in xFit1n1
!
!
!************************************************************
subroutine selectIndividuals(n0, xFitn0, n1, xFitn1) 
    integer, intent(in) :: n0, n1
    type(fitParameters), pointer :: xFitn0(:), xFitn1(:)
    integer :: i, j

    call indexSort(xFitn0%fmax, index, n0)
  
    j = n0 - n1

!write(*, *) "SELECT INDIVIDUALS"
    do i = 1, n1
        j = j + 1
        xFitn1(i) = xFitn0(index( j ) )
!        write(*, *) i, ") f=", xFitn1(i)%fMax, " x=", xFitn1(i)%x        
    end do

!    write(*, *) "selectindividuals, fmax=", xFitn1(n1)%fmax, " x=", xFitn1(n1)%x


end subroutine selectIndividuals







end module optimize_mod
